package asd.td5;

import asd.base.FileDynamique;

/**
 * Une classe pour tous les exercices de feuille 5, sauf l'exercice 4.
 * @author  2002 Marc Gatano
 */
public class ArbreBinaire {

    // attributs

    private Object valeur;       // la valeur du noeud
    private ArbreBinaire gauche; // le sous-arbre gauche
    private ArbreBinaire droite; // le sous-arbre droit

    // constructeurs

    /**
     * Construit un arbre binaire vide.
     */
    public ArbreBinaire() {
        gauche = null;
        droite = null;
    }

    /**
     * Construit un arbre binaire rduit  une feuille
     * de racine x.
     */
    public ArbreBinaire(Object x) {
        valeur = x;
        gauche = new ArbreBinaire();
        droite = new ArbreBinaire();
    }

    /**
     * Construit un arbre binaire de racine x et
     * de sous-arbres gauche et droit g et d.
     */
    public ArbreBinaire(Object x, ArbreBinaire g, ArbreBinaire d) {
        valeur = x;
        gauche = g;
        droite = d;
    }

    // mthodes primitives

    /**
     * Teste si l'arbre est vide.
     */
    public boolean vide() {
        return gauche == null;
    }

    /**
     * Retourne la racine de l'arbre.
     * Prcondition : l'arbre n'est pas vide.
     */
    public Object racine() {
        return valeur;
    }

    /**
     * Retourne le sous-arbre gauche de l'arbre.
     * Prcondition : l'arbre n'est pas vide.
     */
    public ArbreBinaire gauche() {
        return gauche;
    }

    /**
     * Retourne le sous-arbre droit de l'arbre.
     * Prcondition : l'arbre n'est pas vide.
     */
    public ArbreBinaire droite() {
        return droite;
    }

    /**
     * Si l'arbre n'est pas vide, x devient la racine
     * de l'arbre. Si l'arbre est vide, l'arbre devient
     * un arbre binaire rduit  une feuille de racine x.
     */
    public void racine(Object x) {
        if ( vide() ) {
            gauche = new ArbreBinaire();
            droite = new ArbreBinaire();
        }
        valeur = x;
    }

    /**
     * Remplace le sous-arbre gauche de l'arbre par a.
     * Prcondition : l'arbre n'est pas vide.
     */
    public void gauche(ArbreBinaire a) {
        gauche = a;
    }

    /**
     * Remplace le sous-arbre droit de l'arbre par a.
     * Prcondition : l'arbre n'est pas vide.
     */
    public void droite(ArbreBinaire a) {
        droite = a;
    }

    // mthodes non primitives

    /**
     * Affiche l'arbre sous une forme lisible.
     */
    public void afficher() {
        afficher("","");
    }

    /**
     * Affiche l'arbre sous une forme lisible avec r comme prfixe
     * de la chane correspondant  la racine et p comme prfixe
     * des autres chanes.
     */
    private void afficher(String r, String p) {
        if ( vide() ) {
            System.out.println(r);
        }
        else {
            String rs = racine().toString();
            System.out.println(r + rs);
            if ( ! gauche().vide() || ! droite().vide() ) {
                String rr = p + '|' + chaine('_',rs.length()) + ' ';
                droite().afficher(rr, p + '|' + chaine(' ',rs.length() + 1));
                System.out.println(p + '|');
                gauche().afficher(rr, p + chaine(' ',rs.length() + 2));
            }
        }
    }

    /**
     * Retourne une chane de longueur k ne contenant que le
     * caractre c.
     */
    private String chaine(char c, int k) {
        String s = "";
        for ( int i = 0; i < k; i++ ) {
            s += c;
        }
        return s;
    }

    // EXO 1 : mesures

    /**
     * Retourne la taille de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int taille() {
        if ( vide() ) {
            return 0;
        }
        return 1 + gauche().taille() + droite.taille();
    }

    /**
     * Retourne la hauteur de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int hauteur() {
        if ( vide() ) {
            return -1;
        }
        return 1 + Math.max(gauche().hauteur(),droite().hauteur());
    }

    /**
     * Retourne la bassesse de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int bassesse() {
        if ( vide() ) {
            return -1;
        }
        return 1 + Math.min(gauche().bassesse(),droite().bassesse());
    }

    /**
     * Retourne le nombre de feuilles de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int feuilles() {
        if ( vide() ) {
            return 0;
        }
        return fnv();
    }

    /**
     * Retourne le nombre de feuilles de l'arbre.
     * Prcondition : l'arbre est non vide.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    private int fnv(){
        if ( gauche().vide() ) {
            if ( droite().vide() ) {
                return 1;
            }
            return droite().fnv();
        }
        if ( droite().vide() ) {
            return gauche().fnv();
        }
        return gauche().fnv() + droite().fnv();
    }

    // EXO 2 : Base

    /**
     * Retourne la base de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int base() {
        return base(hauteur());
    }

    /**
     * Retourne le nombre d'lments de l'arbre  la profondeur p.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    private int base(int p) {
        if ( vide() ) {
            return 0;
        }
        if ( p == 0 ) {
            return 1;
        }
        return gauche().base(p - 1) + droite().base(p - 1);
    }
        
    // EXO 3 : Gnalogie

    /**
     * Affiche tous les ascendants masculins de racine()  la
     * k-ime gnration.
     * Prcondition : k >= 1
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public void ascendants(int k) {
        if ( ! vide() ) {
            if ( k == 1 ) {
                if ( ! gauche().vide() ) {
                    System.out.println(gauche().racine());
                }
            }
            else {
                gauche().ascendants(k-1);
                droite().ascendants(k-1);
            }
        }
    }

    /**
     * Affiche tous les ascendants masculins de racine() jusqu'
     * la k-ime gnration.
     * Prcondition : k >= 1
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public void tousLesAscendants(int k) {
        if ( ! vide() && k >= 1 ) {
            if ( ! gauche().vide() ) {
                System.out.println(gauche().racine());
            }
            gauche().tousLesAscendants(k-1);
            droite().tousLesAscendants(k-1);
        }
    }

    // EXO 5 : Les deux proprits

    ////////// Penche--Gauche

    /**
     * Teste si l'arbre a la proprit "penche--gauche".
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public boolean pencheAGauche() {
        return pag() != -1;
    }

    /**
     * Si l'arbre a la proprit "penche--gauche", retourne
     * le nombre de ses lments, retourne -1 sinon.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    private int pag() {
        if ( vide() ) {
            return 0;
        }
        int g = gauche().pag();
        if ( g != -1 ) {
            int d = droite().pag();
            if ( d != -1 ) {
                if ( g >= d ) {
                    return g + d + 1;
                }
            }
        }
        return -1;
    }

    ////////// Equilibr

    /**
     * Teste si l'arbre a la proprit "quilibr".
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public boolean equilibre() {
        return eq() != -1;
    }

    /**
     * Si l'arbre A a la proprit "quilibr", retoune
     * sa hauteur plus 1, retourne -1 sinon.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    private int eq() {
        if ( vide() ) {
            return 0;
        }
        int g = gauche().eq();
        if ( g != -1 ) {
            int d = droite().eq();
            if ( d != -1 ) {
                if ( Math.abs(g - d) <= 1 ) {
                    return 1 + Math.max(g,d);
                }
            }
        }
        return -1;
    }

    // EXO 6 : poids

    /**
     * Retourne le poids de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int poids() {
        if ( vide() ) {
            return 0;
        }
        return Math.max(gauche().poids(),droite().poids() + 1);
    }

    /**
     * Transforme l'arbre en un arbre quivalent de poids minimum
     * en utilisant la fonction poids prcdente.
     * Complexit : O(n.lg(n)) o n est le nombre d'lments de l'arbre.
     */
    public void minimiser1() {
        if ( ! vide() ) {
            gauche().minimiser1();
            droite().minimiser1();
            if ( gauche().poids() < droite().poids() ) {
                permuter();
            }
        }
    }

    /**
     * Transforme l'arbre en un arbre quivalent de poids minimum
     * et retourne le poids de l'arbre ainsi transform.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int minimiser2() {
        if ( vide() ) {
            return 0;
        }
        int g = gauche().minimiser2();
        int d = droite().minimiser2();
        if ( g < d ) {
            permuter();
            return d;
        }
        if ( g == d ) {
            return g + 1;
        }
        return g;
    }

    /**
     * Permute les deux sous-arbres gauche et droit de l'arbre.
     */
    private void permuter() {
        ArbreBinaire tmp = gauche();
        gauche(droite());
        droite(tmp);
    }

    // EXO 7 : Harmonieux
    
    /**
     * Teste si l'arbre a la proprit "harmonieux".
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public boolean harmonieux() {
        return harmo().estHarmonieux;
    }

    /**
     * Retourne un objet Harmonieux qui dit si l'arbre est harmonieux.
     * Dans l'affirmative, cet objet contient la hauteur et la bassesse
     * de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    private Harmonieux harmo() {
        if ( vide() ) {
            return new Harmonieux(-1,-1);
        }
        Harmonieux g = gauche().harmo();
        if ( g.estHarmonieux ) {
            Harmonieux d = droite().harmo();
            if ( d.estHarmonieux ) {
                int h = Math.max(g.hauteur,d.hauteur)   + 1;
                int b = Math.min(g.bassesse,d.bassesse) + 1;
                if ( h <= 2*b ) {
                    return new Harmonieux(h,b);
                }
            }
        }
        return new Harmonieux();
    }

    /**
     * Une classe interne pour la proprit "harmonieux".
     * Tout est en O(1).
     */
    private class Harmonieux {

        boolean estHarmonieux; // indique si l'arbre est harmonieux
        int hauteur;           // la hauteur de l'arbre
        int bassesse;          // la bassesse

        /**
         * Construit un objet Harmonieux qui indique que
         * l'arbre n'est pas harmonieux.
         */
        public Harmonieux() {
            estHarmonieux = false;
        }

        /**
         * Construit un objet Harmonieux qui indique que
         * l'arbre est harmonieux et qu'il possde la
         * hauteur h et la bassesse b.
         */
        public Harmonieux(int h, int b) {
            estHarmonieux = true;
            hauteur = h;
            bassesse = b;
        }
    }

    // EXO 8 : Parcours par niveaux

    /**
     * Affiche sur le flux de sortie tous les lments de l'arbre
     * dans l'ordre "par niveau". Ce parcours est un parcours en
     * largeur qui diffrencie les profondeurs des lments.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public void parcoursNiveau() {
        FileDynamique F = new FileDynamique();
        F.ajouter(this);
        while ( ! F.vide() ) {
            int max = F.longueur();
            for ( int i = 0; i < max; i++ ) {
                ArbreBinaire A = (ArbreBinaire) F.retirer();
                System.out.print(A.racine() + " ");
                if ( ! A.gauche().vide() ) {
                    F.ajouter(A.gauche());
                }
                if ( ! A.droite().vide() ) {
                    F.ajouter(A.droite());
                }
            }
            System.out.println();
        }
    }

    // EXO 9 : Largeur

    /**
     * Calcule la largeur de l'arbre, soit le nombre maximum d'lments
     *  la mme profondeur.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int largeur() {
        FileDynamique F = new FileDynamique();
        F.ajouter(this);
        int largeur = 0;
        while ( ! F.vide() ) {
            if ( F.longueur() > largeur ) {
                largeur = F.longueur();
            }
            int max = F.longueur();
            for ( int i = 0; i < max; i++ ) {
                ArbreBinaire A = (ArbreBinaire) F.retirer();
                if ( ! A.gauche().vide() ) {
                    F.ajouter(A.gauche());
                }
                if ( ! A.droite().vide() ) {
                    F.ajouter(A.droite());
                }
            }
        }
        return largeur;
    }

    // EXO 10 : Gnalogie (2)

    /**
     * Affiche tous les ascendants masculins de racine()  la
     * k-ime gnration.
     * Prcondition : k >= 1
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public void ascendants2(int k) {
        FileDynamique F = new FileDynamique();
        F.ajouter(this);
        for ( int i = 0; i < k - 1; i++ ) {
            int max = F.longueur();
            for ( int j = 0; j < max; j++ ) {
                ArbreBinaire A = (ArbreBinaire) F.retirer();
                if ( ! A.gauche().vide() ) {
                    F.ajouter(A.gauche());
                }
                if ( ! A.droite().vide() ) {
                    F.ajouter(A.droite());
                }
            }
        }
        while ( ! F.vide() ) {
            ArbreBinaire A = (ArbreBinaire) F.retirer();
            if ( ! A.gauche().vide() ) {
                System.out.println(A.gauche().racine());
            }
        }
    }

    /**
     * Affiche tous les ascendants masculins de racine() jusqu'
     * la k-ime gnration.
     * Prcondition : k >= 1
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public void tousLesAscendants2(int k) {
        FileDynamique F = new FileDynamique();
        F.ajouter(this);
        for ( int i = 0; i < k - 1; i++ ) {
            int max = F.longueur();
            for ( int j = 0; j < max; j++ ) {
                ArbreBinaire A = (ArbreBinaire) F.retirer();
                if ( ! A.gauche().vide() ) {
                    System.out.println(A.gauche().racine());
                    F.ajouter(A.gauche());
                }
                if ( ! A.droite().vide() ) {
                    F.ajouter(A.droite());
                }
            }
        }
        while ( ! F.vide() ) {
            ArbreBinaire A = (ArbreBinaire) F.retirer();
            if ( ! A.gauche().vide() ) {
                System.out.println(A.gauche().racine());
            }
        }
    }
}
